/* dlvhex -- Answer-Set Programming with external interfaces.
 * Copyright (C) 2005-2007 Roman Schindlauer
 * Copyright (C) 2006-2015 Thomas Krennwallner
 * Copyright (C) 2009-2016 Peter Schüller
 * Copyright (C) 2011-2016 Christoph Redl
 * Copyright (C) 2015-2016 Tobias Kaminski
 * Copyright (C) 2015-2016 Antonius Weinzierl
 *
 * This file is part of dlvhex.
 *
 * dlvhex is free software; you can redistribute it and/or modify it
 * under the terms of the GNU Lesser General Public License as
 * published by the Free Software Foundation; either version 2.1 of
 * the License, or (at your option) any later version.
 *
 * dlvhex is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with dlvhex; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA
 * 02110-1301 USA.
 */

/**
 * @file   FLPModelGeneratorBase.cpp
 * @author Peter Schueller <ps@kr.tuwien.ac.at>
 * @author Chrisoph Redl <redl@kr.tuwien.ac.at>
 *
 * @brief  Base class for model generators using FLP reduct.
 */

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif                           // HAVE_CONFIG_H

#include "dlvhex2/FLPModelGeneratorBase.h"
#include "dlvhex2/Printer.h"
#include "dlvhex2/ProgramCtx.h"
#include "dlvhex2/LiberalSafetyChecker.h"

#include "dlvhex2/CDNLSolver.h"
#include "dlvhex2/ClaspSolver.h"
#include "dlvhex2/SATSolver.h"
#include "dlvhex2/Printer.h"

#include <boost/graph/graph_traits.hpp>
#include <boost/graph/adjacency_list.hpp>
#include <boost/graph/breadth_first_search.hpp>
#include <boost/graph/visitors.hpp>

#include <fstream>

DLVHEX_NAMESPACE_BEGIN

FLPModelGeneratorFactoryBase::FLPModelGeneratorFactoryBase(
ProgramCtx& ctx):
ctx(ctx), reg(ctx.registry())
{
    gpMask.setRegistry(reg);
    gnMask.setRegistry(reg);
    fMask.setRegistry(reg);
}


/**
 * go through all rules with external atoms
 * for each such rule and each inner eatom in the body:
 * * collect all variables in the eatom (input and output)
 * * collect all positive non-external predicates in the rule body containing these variables
 * * build rule <aux_ext_eatompos>(<all variables>) v <aux_ext_eatomneg>(<all variables>) :- <all bodies>
 * * store into gidb
 */
void FLPModelGeneratorFactoryBase::createEatomGuessingRules(const ProgramCtx& ctx)
{
    std::set<ID> innerEatomsSet(innerEatoms.begin(), innerEatoms.end());
    assert((innerEatomsSet.empty() ||
        (!innerEatomsSet.begin()->isLiteral() && innerEatomsSet.begin()->isExternalAtom())) &&
        "we don't want literals here, we want external atoms");

    DBGLOG_SCOPE(DBG,"cEAGR",false);
    BOOST_FOREACH(ID rid, idb) {
        // skip rules without external atoms
        if( !rid.doesRuleContainExtatoms() )
            continue;

        // do not guess external atoms in auxiliary input rules
        // because those rules may not contain all relevant body atoms which provide grounding
        //    if ( rid.isExternalInputAuxiliary() )
        //      continue;

        const Rule& r = reg->rules.getByID(rid);
        DBGLOG(DBG,"processing rule with external atom(s): " << printToString<RawPrinter>(rid, reg) <<
            " (rid " << rid << "r " << r << ")");

        BOOST_FOREACH(ID lit, r.body) {
            // skip atoms that are not external atoms
            if( !lit.isExternalAtom() )
                continue;

            if( innerEatomsSet.count(ID::atomFromLiteral(lit)) == 0 )
                continue;

            gidb.push_back(createEatomGuessingRule(ctx, rid, lit));
        }
    }
}


ID FLPModelGeneratorFactoryBase::createEatomGuessingRule(const ProgramCtx& ctx, ID rid, ID lit)
{
    const Rule& r = reg->rules.getByID(rid);
    const ExternalAtom& eatom = reg->eatoms.getByID(lit);
    DBGLOG(DBG,"processing external atom " << printToString<RawPrinter>(lit, reg) <<
        " (lit " << lit << " eatom " << eatom << ")");
    DBGLOG_INDENT(DBG);

    // prepare replacement atom
    OrdinaryAtom replacement(
        ID::MAINKIND_ATOM | ID::PROPERTY_AUX | ID::PROPERTY_EXTERNALAUX);

    // tuple: (replacement_predicate, inputs_as_in_inputtuple*, outputs*)
    // (build up incrementally)
    ID pospredicate = reg->getAuxiliaryConstantSymbol('r', eatom.predicate);
    ID negpredicate = reg->getAuxiliaryConstantSymbol('n', eatom.predicate);

    replacement.tuple.push_back(pospredicate);
    gpMask.addPredicate(pospredicate);
    gnMask.addPredicate(negpredicate);

    if (ctx.config.getOption("IncludeAuxInputInAuxiliaries") && eatom.auxInputPredicate != ID_FAIL)
        replacement.tuple.push_back(eatom.auxInputPredicate);

    // build (nonground) replacement and harvest all variables
    std::set<ID> variables;
    BOOST_FOREACH(ID inp, eatom.inputs) {
        replacement.tuple.push_back(inp);
        if( inp.isVariableTerm() )
            variables.insert(inp);
    }
    BOOST_FOREACH(ID outp, eatom.tuple) {
        replacement.tuple.push_back(outp);
        if( outp.isVariableTerm() )
            variables.insert(outp);
    }
    DBGLOG(DBG,"found set of variables: " << printManyToString<RawPrinter>(Tuple(variables.begin(), variables.end()), ",", reg));

    // groundness of replacement predicate
    ID posreplacement;
    ID negreplacement;
    if( variables.empty() ) {
        replacement.kind |= ID::SUBKIND_ATOM_ORDINARYG;
        posreplacement = reg->storeOrdinaryGAtom(replacement);
        replacement.tuple[0] = negpredicate;
        negreplacement = reg->storeOrdinaryGAtom(replacement);
    }
    else {
        replacement.kind |= ID::SUBKIND_ATOM_ORDINARYN;
        posreplacement = reg->storeOrdinaryNAtom(replacement);
        replacement.tuple[0] = negpredicate;
        negreplacement = reg->storeOrdinaryNAtom(replacement);
    }
    DBGLOG(DBG,"registered posreplacement " << printToString<RawPrinter>(posreplacement, reg) <<
        " and negreplacement " << printToString<RawPrinter>(negreplacement, reg));

    // create rule head
    Rule guessingrule(ID::MAINKIND_RULE | ID::SUBKIND_RULE_REGULAR |
        ID::PROPERTY_AUX | ID::PROPERTY_RULE_DISJ);
    guessingrule.head.push_back(posreplacement);
    guessingrule.head.push_back(negreplacement);

    // create rule body (if there are variables that need to be grounded)
    if( !variables.empty() ) {
        // harvest all positive ordinary nonground atoms
        // "grounding the variables" (i.e., those that contain them)
        BOOST_FOREACH(ID lit, r.body) {
            if( lit.isNaf() ||
                lit.isExternalAtom() )
                continue;

            bool use = false;
            if( lit.isOrdinaryNongroundAtom() ) {
                const OrdinaryAtom& oatom = reg->onatoms.getByID(lit);
                // look if this atom grounds any variables we need
                BOOST_FOREACH(ID term, oatom.tuple) {
                    if( term.isVariableTerm() &&
                    (variables.find(term) != variables.end()) ) {
                        use = true;
                        break;
                    }
                }
            }
            else if( lit.isBuiltinAtom() ) {
                const BuiltinAtom& biatom = reg->batoms.getByID(lit);
                // !=, <, >, <= and >= cannot provide grounding
                if (biatom.tuple[0].address == ID::TERM_BUILTIN_NE || biatom.tuple[0].address == ID::TERM_BUILTIN_LT || biatom.tuple[0].address == ID::TERM_BUILTIN_LE ||
                    biatom.tuple[0].address == ID::TERM_BUILTIN_GT || biatom.tuple[0].address == ID::TERM_BUILTIN_GE) continue;

                // look if this atom grounds any variables we need
                BOOST_FOREACH(ID term, biatom.tuple) {
                    if( term.isVariableTerm() &&
                    (variables.find(term) != variables.end()) ) {
                        use = true;
                        break;
                    }
                }
            }
            else if (lit.isAggregateAtom()) {
                // take the aggregate iff it defines a variable to be grounded
                const AggregateAtom& aatom = reg->aatoms.getByID(lit);
                DBGLOG(DBG, "Checking if aggregate is included in guessing rule");
                if ( (aatom.tuple[1].address == ID::TERM_BUILTIN_EQ && aatom.tuple[0].isVariableTerm() && variables.find(aatom.tuple[0]) != variables.end()) || (aatom.tuple[3].address == ID::TERM_BUILTIN_EQ && aatom.tuple[4].isVariableTerm() && variables.find(aatom.tuple[4]) != variables.end())) {
                    DBGLOG(DBG, "Aggregate is included in guessing rule");
                    use = true;
                }
            }

            if( use ) {
                guessingrule.body.push_back(lit);
            }
        }
    }

    // the auxiliary input also provides grounding (potentially)
    if (eatom.auxInputPredicate != ID_FAIL) {
        DBGLOG(DBG, "Adding auxiliary input predicate " << printToString<RawPrinter>(eatom.auxInputPredicate,reg) << " to guessing rule");
        OrdinaryAtom auxinput(ID::MAINKIND_ATOM | ID::SUBKIND_ATOM_ORDINARYN | ID::PROPERTY_AUX | ID::PROPERTY_EXTERNALINPUTAUX);
        auxinput.tuple.push_back(eatom.auxInputPredicate);
        // resize to hold input predicate and all aux input variables
        auxinput.tuple.resize(eatom.auxInputMapping.size()+1,ID_FAIL);
        // now assign correct variables from inputs to aux inputs
        unsigned at = 1;
        for(ExternalAtom::AuxInputMapping::const_iterator itaim = eatom.auxInputMapping.begin();
        itaim != eatom.auxInputMapping.end(); ++itaim, ++at) {
            typedef std::list<unsigned> UList;
            const UList& varplaces = *itaim;
            ID current = ID_FAIL;
            assert(!varplaces.empty() && "cannot have empty variable mapping");
            for(UList::const_iterator it = varplaces.begin();
            it != varplaces.end(); ++it) {
                if( it == varplaces.begin() ) {
                    // set the variable
                    current = eatom.inputs[*it];
                }
                else {
                    // verify the variable
                    assert(current == eatom.inputs[*it] && "something went wrong with auxInputMapping!");
                }
            }
            auxinput.tuple[at] = current;
        }
        ID aiid = reg->storeOrdinaryNAtom(auxinput);
        DBGLOG(DBG,"created auxiliary grounding predicate " << printToString<RawPrinter>(aiid,reg) << " which got id " << aiid);
        guessingrule.body.push_back(ID::posLiteralFromAtom(aiid));
    }

    // store rule
    ID gid = reg->storeRule(guessingrule);
    DBGLOG(DBG,"stored guessingrule " << printToString<RawPrinter>(gid, reg) << " which got id " << gid);
    return gid;
}


/**
 * for each rule in xidb
 * * keep disjunctive facts: copy ID to xidbflphead and xidbflpbody
 * * (note: nondisjunctive facts are stored in edb)
 * * for all other rules:
 * * collect all variables in the body (which means also all variables in the head)
 * * create ground or nonground flp replacement atom containing all variables
 * * create rule <flpreplacement>(<allvariables>) :- <body> and store in xidbflphead
 * * create rule <head> :- <flpreplacement>(<allvariables>), <body> and store in xidbflpbody
 */

void FLPModelGeneratorFactoryBase::createFLPRules()
{
    DBGLOG_SCOPE(DBG,"cFLPR",false);
    BOOST_FOREACH(ID rid, xidb) {
        const Rule& r = reg->rules.getByID(rid);
        DBGLOG(DBG,"processing rule " << rid << " " << r);
        if( r.body.empty() ) {
            // keep disjunctive facts as they are
            xidbflphead.push_back(rid);
            xidbflpbody.push_back(rid);
        }
        else if( rid.isConstraint() || rid.isRegularRule() ) {
            // collect all variables
            std::set<ID> variables;
            BOOST_FOREACH(ID lit, r.body) {
                assert(!lit.isExternalAtom() && "in xidb there must not be external atoms left");
                WARNING("TODO factorize get all (free) variables from entity")
                // from ground literals we don't need variables
                    if( lit.isOrdinaryGroundAtom() )
                    continue;

                if( lit.isOrdinaryNongroundAtom() ) {
                    const OrdinaryAtom& onatom = reg->onatoms.getByID(lit);
                    BOOST_FOREACH(ID idt, onatom.tuple) {
                        if( idt.isVariableTerm() )
                            variables.insert(idt);
                    }
                }
                else if( lit.isBuiltinAtom() ) {
                    const BuiltinAtom& batom = reg->batoms.getByID(lit);
                    BOOST_FOREACH(ID idt, batom.tuple) {
                        if( idt.isVariableTerm() )
                            variables.insert(idt);
                    }
                }
                else if( lit.isAggregateAtom() ) {
                    // use boundaries as variables, variables in aggregate are free or (automatically) bound by context in rule
                    const AggregateAtom& aatom = reg->aatoms.getByID(lit);
                    if( aatom.tuple[0] != ID_FAIL && aatom.tuple[0].isVariableTerm() )
                        variables.insert(aatom.tuple[0]);
                    if( aatom.tuple[4] != ID_FAIL && aatom.tuple[4].isVariableTerm() )
                        variables.insert(aatom.tuple[4]);
                }
                else {
                    LOG(ERROR,"encountered literal " << lit << " in FLP check, don't know what to do about it");
                    throw FatalError("TODO: think about how to treat other types of atoms in FLP check");
                }
            }
            DBGLOG(DBG,"collected variables " << printset(variables));

            // prepare replacement atom
            OrdinaryAtom replacement(
                ID::MAINKIND_ATOM | ID::PROPERTY_AUX | ID::PROPERTY_FLPAUX);

            // tuple: (replacement_predicate, variables*)
            ID flppredicate = reg->getAuxiliaryConstantSymbol('f', rid);
            replacement.tuple.push_back(flppredicate);
            fMask.addPredicate(flppredicate);

            // groundness of replacement predicate
            ID fid;
            if( variables.empty() ) {
                replacement.kind |= ID::SUBKIND_ATOM_ORDINARYG;
                fid = reg->storeOrdinaryGAtom(replacement);
            }
            else {
                replacement.kind |= ID::SUBKIND_ATOM_ORDINARYN;
                replacement.tuple.insert(replacement.tuple.end(),
                    variables.begin(), variables.end());
                fid = reg->storeOrdinaryNAtom(replacement);
            }
            DBGLOG(DBG,"registered flp replacement " << replacement <<
                " with fid " << fid);

            // create rules
            Rule rflphead(
                ID::MAINKIND_RULE | ID::SUBKIND_RULE_REGULAR | ID::PROPERTY_AUX);
            rflphead.head.push_back(fid);
            rflphead.body = r.body;

            // kind will be overwritten
            Rule rflpbody(ID::MAINKIND_RULE | ID::SUBKIND_RULE_REGULAR);

            // Note: EA-aux input rules MUST NOT be shifted! This could eliminate models of the reduct
            if( r.isEAAuxInputRule() || ctx.config.getOption("ExplicitFLPUnshift") == 1 ) {
                // original set of rules
                IDKind kind = ID::MAINKIND_RULE | ID::PROPERTY_AUX;
                if (r.head.size() == 0) {
                    kind |= ID::SUBKIND_RULE_CONSTRAINT;
                }
                else {
                    kind |= ID::SUBKIND_RULE_REGULAR;
                }
                rflpbody.kind = kind;
                rflpbody.head = r.head;
                if( rflpbody.head.size() > 1 )
                    rflpbody.kind |= ID::PROPERTY_RULE_DISJ;
                rflpbody.body = r.body;
                rflpbody.body.push_back(fid);
            }
            else {
                // optimized set of rules
                // another encoding which is more efficient on some examples:
                IDKind kind = ID::MAINKIND_RULE | ID::SUBKIND_RULE_CONSTRAINT | ID::PROPERTY_AUX;
                rflpbody.kind = kind | ID::SUBKIND_RULE_CONSTRAINT;
                rflpbody.body = r.body;
                rflpbody.body.push_back(fid);
                BOOST_FOREACH (ID h, r.head) {
                    rflpbody.body.push_back(ID::literalFromAtom(h, true));
                }
            }

            // store rules
            ID fheadrid = reg->storeRule(rflphead);
            xidbflphead.push_back(fheadrid);
            ID fbodyrid = reg->storeRule(rflpbody);
            xidbflpbody.push_back(fbodyrid);

            #ifndef NDEBUG
            {
                std::stringstream s;
                RawPrinter p(s, reg);
                p.print(fheadrid);
                s << " and ";
                p.print(fbodyrid);
                DBGLOG(DBG,"stored flphead rule " << rflphead << " which got id " << fheadrid);
                DBGLOG(DBG,"stored flpbody rule " << rflpbody << " which got id " << fbodyrid);
                DBGLOG(DBG,"rules are " << s.str());
            }
            #endif
        }
        else {
            LOG(ERROR,"got weak rule " << r << " in guess and check model generator, don't know what to do about it");
            throw FatalError("TODO: think about weak rules in G&C MG");
        }
    }
}


//
// FLPModelGeneratorBase
//

FLPModelGeneratorBase::FLPModelGeneratorBase(
FLPModelGeneratorFactoryBase& _factory, InterpretationConstPtr input):
BaseModelGenerator(input),
factory(_factory),
annotatedGroundProgram(_factory.ctx, _factory.innerEatoms)
{
}


bool FLPModelGeneratorBase::isCompatibleSet(
InterpretationConstPtr candidateCompatibleSet,
InterpretationConstPtr postprocessedInput,
ProgramCtx& ctx,
SimpleNogoodContainerPtr nc)
{
    RegistryPtr& reg = factory.reg;
    PredicateMask& gpMask = factory.gpMask;
    PredicateMask& gnMask = factory.gnMask;

    // project to pos and neg eatom replacements for validation
    InterpretationPtr projint(new Interpretation(reg));
    projint->getStorage() = candidateCompatibleSet->getStorage() - postprocessedInput->getStorage();

    gpMask.updateMask();
    InterpretationPtr projectedModelCandidate_pos(new Interpretation(reg));
    projectedModelCandidate_pos->getStorage() = projint->getStorage() & gpMask.mask()->getStorage();
    InterpretationPtr projectedModelCandidate_pos_val(new Interpretation(reg));
    projectedModelCandidate_pos_val->getStorage() = projectedModelCandidate_pos->getStorage();
    DBGLOG(DBG,"projected positive guess: " << *projectedModelCandidate_pos);

    gnMask.updateMask();
    InterpretationPtr projectedModelCandidate_neg(new Interpretation(reg));
    projectedModelCandidate_neg->getStorage() = projint->getStorage() & gnMask.mask()->getStorage();
    DBGLOG(DBG,"projected negative guess: " << *projectedModelCandidate_neg);

    // verify whether correct eatoms where guessed true
    // this callback checks if a positive eatom result was guessed as negative
    // -> in this case it aborts
    // this callback resets all positive bits it encounters
    // -> if the positive interpretation is all-zeroes at the end,
    //    the guess was correct
    VerifyExternalAnswerAgainstPosNegGuessInterpretationCB cb(
        projectedModelCandidate_pos_val, projectedModelCandidate_neg);

    // we might need edb facts here
    // (dependencies to edb are not modelled in the dependency graph)
    // therefore we did not mask the guess program before
    if( !evaluateExternalAtoms(
        ctx, factory.innerEatoms,
        candidateCompatibleSet, cb,
    ctx.config.getOption("ExternalLearning") ? nc : SimpleNogoodContainerPtr())) {
        return false;
    }

    // check if we guessed too many true atoms
    if (projectedModelCandidate_pos_val->getStorage().count() > 0) {
        return false;
    }
    return true;
}


Nogood FLPModelGeneratorBase::getFLPNogood(
ProgramCtx& ctx,
const OrdinaryASPProgram& groundProgram,
InterpretationConstPtr compatibleSet,
InterpretationConstPtr projectedCompatibleSet,
InterpretationConstPtr smallerFLPModel
)
{

    RegistryPtr reg = factory.reg;

    Nogood ng;

    // for each rule with unsatisfied body
    BOOST_FOREACH (ID ruleId, groundProgram.idb) {
        const Rule& rule = reg->rules.getByID(ruleId);
        BOOST_FOREACH (ID b, rule.body) {
            if (compatibleSet->getFact(b.address) != !b.isNaf()) {
                // take an unsatisfied body literal
                ng.insert(NogoodContainer::createLiteral(b.address, compatibleSet->getFact(b.address)));
                break;
            }
        }
    }

    // add the smaller FLP model
    bm::bvector<>::enumerator en = smallerFLPModel->getStorage().first();
    bm::bvector<>::enumerator en_end = smallerFLPModel->getStorage().end();
    while (en < en_end) {
        ng.insert(NogoodContainer::createLiteral(*en));
        en++;
    }

    // add one atom which is in the compatible set but not in the flp model
    en = projectedCompatibleSet->getStorage().first();
    en_end = projectedCompatibleSet->getStorage().end();
    while (en < en_end) {
        if (!smallerFLPModel->getFact(*en)) {
            ng.insert(NogoodContainer::createLiteral(*en));
            break;
        }
        en++;
    }

    DBGLOG(DBG, "Constructed FLP nogood " << ng);

    return ng;
}


WARNING("TODO could we move shadow predicates and mappings and rules to factory?")
void FLPModelGeneratorBase::computeShadowAndUnfoundedPredicates(
RegistryPtr reg,
InterpretationConstPtr edb,
const std::vector<ID>& idb,
std::map<ID, std::pair<int, ID> >& shadowPredicates,
std::map<ID, std::pair<int, ID> >& unfoundedPredicates,
std::string& shadowPostfix,
std::string& unfoundedPostfix)
{

    // collect predicates
    std::set<std::pair<int, ID> > preds;

    // edb
    bm::bvector<>::enumerator en = edb->getStorage().first();
    bm::bvector<>::enumerator en_end = edb->getStorage().end();
    while (en < en_end) {
        if (!reg->ogatoms.getIDByAddress(*en).isExternalAuxiliary() && !reg->ogatoms.getIDByAddress(*en).isFLPAuxiliary()) {
            const OrdinaryAtom& atom = reg->ogatoms.getByAddress(*en);
            preds.insert(std::pair<int, ID>(atom.tuple.size() - 1, atom.tuple.front()));
        }
        en++;
    }

    // idb
    BOOST_FOREACH (ID rid, idb) {
        const Rule& r = reg->rules.getByID(rid);
        BOOST_FOREACH (ID h, r.head) {
            if (!h.isExternalAuxiliary() && !h.isFLPAuxiliary()) {
                const OrdinaryAtom& atom = h.isOrdinaryGroundAtom() ? reg->ogatoms.getByID(h) : reg->onatoms.getByID(h);
                preds.insert(std::pair<int, ID>(atom.tuple.size() - 1, atom.tuple.front()));
            }
        }
        BOOST_FOREACH (ID b, r.body) {
            if (b.isOrdinaryAtom() && !b.isExternalAuxiliary() && !b.isFLPAuxiliary()) {
                const OrdinaryAtom& atom = b.isOrdinaryGroundAtom() ? reg->ogatoms.getByID(b) : reg->onatoms.getByID(b);
                preds.insert(std::pair<int, ID>(atom.tuple.size() - 1, atom.tuple.front()));
            }
        }
    }

    // create unique predicate suffix for shadow predicates
    // (must not start with _ as it will be used by itself and
    // constants starting with _ are forbidden in dlv as they are not c-identifiers)
    shadowPostfix = "shadow";
    int idx = 0;
    bool clash;
    do {
        clash = false;

        // check if the current postfix clashes with any of the predicates
        typedef std::pair<int, ID> Pair;
        BOOST_FOREACH (Pair p, preds) {
            std::string currentPred = reg->terms.getByID(p.second).getUnquotedString();
                                 // currentPred is at least as long as shadowPostfix
            if (shadowPostfix.length() <= currentPred.length() &&
                                 // postfixes coincide
            currentPred.substr(currentPred.length() - shadowPostfix.length()) == shadowPostfix) {
                clash = true;
                break;
            }
        }
        std::stringstream ss;
        ss << "shadow" << idx++;
    }while(clash);

    // create shadow predicates
    typedef std::pair<int, ID> Pair;
    BOOST_FOREACH (Pair p, preds) {
        Term shadowTerm(ID::MAINKIND_TERM | ID::SUBKIND_TERM_CONSTANT, reg->terms.getByID(p.second).getUnquotedString() + shadowPostfix);
        ID shadowID = reg->storeTerm(shadowTerm);
        shadowPredicates[p.second] = Pair(p.first, shadowID);
        DBGLOG(DBG, "Predicate " << reg->terms.getByID(p.second).getUnquotedString() << " [" << p.second << "] has shadow predicate " <<
            reg->terms.getByID(p.second).getUnquotedString() + shadowPostfix << " [" << shadowID << "]");
    }

    // create unfounded predicate suffix for shadow predicates
    unfoundedPostfix = "_unfounded";
    idx = 0;
    do {
        clash = false;

        // check if the current postfix clashes with any of the predicates
        typedef std::pair<int, ID> Pair;
        BOOST_FOREACH (Pair p, preds) {
            std::string currentPred = reg->terms.getByID(p.second).getUnquotedString();
                                 // currentPred is at least as long as shadowPostfix
            if (unfoundedPostfix.length() <= currentPred.length() &&
                                 // postfixes coincide
            currentPred.substr(currentPred.length() - unfoundedPostfix.length()) == unfoundedPostfix) {
                clash = true;
                break;
            }
        }
        std::stringstream ss;
        ss << "unfounded" << idx++;
    }while(clash);

    // create unfounded predicates
    BOOST_FOREACH (Pair p, preds) {
        Term unfoundedTerm(ID::MAINKIND_TERM | ID::SUBKIND_TERM_CONSTANT, reg->terms.getByID(p.second).getUnquotedString() + unfoundedPostfix);
        ID unfoundedID = reg->storeTerm(unfoundedTerm);
        unfoundedPredicates[p.second] = Pair(p.first, unfoundedID);
        DBGLOG(DBG, "Predicate " << reg->terms.getByID(p.second).getUnquotedString() << " [" << p.second << "] has unfounded predicate " <<
            reg->terms.getByID(p.second).getUnquotedString() + unfoundedPostfix << " [" << unfoundedID << "]");
    }
}


void FLPModelGeneratorBase::addShadowInterpretation(
RegistryPtr reg,
std::map<ID, std::pair<int, ID> >& shadowPredicates,
InterpretationConstPtr input,
InterpretationPtr output)
{

    bm::bvector<>::enumerator en = input->getStorage().first();
    bm::bvector<>::enumerator en_end = input->getStorage().end();
    while (en < en_end) {
        OrdinaryAtom atom = reg->ogatoms.getByID(ID(ID::MAINKIND_ATOM | ID::SUBKIND_ATOM_ORDINARYG, *en));
        if (shadowPredicates.find(atom.tuple[0]) != shadowPredicates.end()) {
            atom.tuple[0] = shadowPredicates[atom.tuple[0]].second;
            output->setFact(reg->storeOrdinaryGAtom(atom).address);
        }
        en++;
    }
}


void FLPModelGeneratorBase::createMinimalityRules(
RegistryPtr reg,
std::map<ID, std::pair<int, ID> >& shadowPredicates,
std::string& shadowPostfix,
std::vector<ID>& idb)
{

    // construct a propositional atom which does neither occur in the input program nor as a shadow predicate
    // for this purpose we use the shadowPostfix alone:
    // - it cannot be used by the input program (otherwise it would not be a postfix)
    // - it cannot be used by the shadow atoms (otherwise an input atom would be the empty string, which is not possible)
    Term smallerTerm(ID::MAINKIND_TERM | ID::SUBKIND_TERM_CONSTANT, shadowPostfix);
    OrdinaryAtom smallerAtom(ID::MAINKIND_ATOM | ID::SUBKIND_ATOM_ORDINARYG);
    smallerAtom.tuple.push_back(reg->storeTerm(smallerTerm));
    ID smallerAtomID = reg->storeOrdinaryGAtom(smallerAtom);

    typedef std::pair<ID, std::pair<int, ID> > Pair;
    BOOST_FOREACH (Pair p, shadowPredicates) {
        OrdinaryAtom atom(ID::MAINKIND_ATOM);
        if (p.second.first == 0) atom.kind |= ID::SUBKIND_ATOM_ORDINARYG; else atom.kind |= ID::SUBKIND_ATOM_ORDINARYN;
        if (p.first.isAuxiliary()) atom.kind |= ID::PROPERTY_AUX;
        atom.tuple.push_back(p.first);
        for (int i = 0; i < p.second.first; ++i) {
            std::stringstream var;
            var << "X" << i;
            atom.tuple.push_back(reg->storeVariableTerm(var.str()));
        }

        // store original atom
        ID origID;
        if (p.second.first == 0) {
            origID = reg->storeOrdinaryGAtom(atom);
        }
        else {
            origID = reg->storeOrdinaryNAtom(atom);
        }

        // store shadow atom
        atom.kind = ID::MAINKIND_ATOM;
        if (p.second.first == 0) atom.kind |= ID::SUBKIND_ATOM_ORDINARYG; else atom.kind |= ID::SUBKIND_ATOM_ORDINARYN;
        atom.tuple[0] = p.second.second;
        ID shadowID;
        if (p.second.first == 0) {
            shadowID = reg->storeOrdinaryGAtom(atom);
        }
        else {
            shadowID = reg->storeOrdinaryNAtom(atom);
        }
        DBGLOG(DBG, "Using shadow predicate for " << p.first << " which is " << p.second.second);

        // an atom must not be true if the shadow atom is false because the computed interpretation must be a subset of the shadow interpretation
        {
            // construct rule   :- a, not a_shadow   to ensure that the models are (not necessarily proper) subsets of the shadow model
            Rule r(ID::MAINKIND_RULE | ID::SUBKIND_RULE_CONSTRAINT);
            ID id = origID;
            r.body.push_back(id);
            id = ID(ID::MAINKIND_LITERAL | ID::NAF_MASK | (shadowID.kind & ID::SUBKIND_MASK), shadowID.address);
            r.body.push_back(id);
            idb.push_back(reg->storeRule(r));
        }

        // but we want a proper subset, so add a rule   smaller :- a_shadow, not a
        // an atom must not be true if the shadow atom is false because the computed interpretation must be a subset of the shadow interpretation
        {
            // construct rule   :- a, not a_shadow   to ensure that the models are (not necessarily proper) subsets of the shadow model
            Rule r(ID::MAINKIND_RULE | ID::SUBKIND_RULE_REGULAR);
            ID id = smallerAtomID;
            r.head.push_back(id);
            id = ID(ID::MAINKIND_LITERAL | ID::NAF_MASK | (origID.kind & ID::SUBKIND_MASK), origID.address);
            r.body.push_back(id);
            r.body.push_back(shadowID);
            idb.push_back(reg->storeRule(r));
        }
    }

    // construct a rule   :- not smaller   to restrict the search space to proper submodels of the shadow model
    Rule r(ID::MAINKIND_RULE | ID::SUBKIND_RULE_CONSTRAINT);
    r.body.push_back(ID(ID::MAINKIND_LITERAL | ID::SUBKIND_ATOM_ORDINARYG | ID::NAF_MASK, smallerAtomID.address));
    idb.push_back(reg->storeRule(r));
}


void FLPModelGeneratorBase::createFoundingRules(
RegistryPtr reg,
std::map<ID, std::pair<int, ID> >& shadowPredicates,
std::map<ID, std::pair<int, ID> >& unfoundedPredicates,
std::vector<ID>& idb)
{

    // We want to compute a _model_ of the reduct rather than an _answer set_,
    // i.e., atoms are allowed to be _not_ founded.
    // For this we introduce for each n-ary shadow predicate
    //	ps(X1, ..., Xn)
    // a rule
    //	p(X1, ..., Xn) v p_unfounded(X1, ..., Xn) :- ps(X1, ..., Xn)
    // which can be used to found an atom.
    // (p_unfounded(X1, ..., Xn) encodes that the atom is not artificially founded)

    typedef std::pair<ID, std::pair<int, ID> > Pair;
    BOOST_FOREACH (Pair p, shadowPredicates) {
        OrdinaryAtom atom(ID::MAINKIND_ATOM);
        if (p.second.first == 0) atom.kind |= ID::SUBKIND_ATOM_ORDINARYG; else atom.kind |= ID::SUBKIND_ATOM_ORDINARYN;
        if (p.first.isAuxiliary()) atom.kind |= ID::PROPERTY_AUX;
        atom.tuple.push_back(p.first);
        for (int i = 0; i < p.second.first; ++i) {
            std::stringstream var;
            var << "X" << i;
            atom.tuple.push_back(reg->storeVariableTerm(var.str()));
        }

        // store original atom
        ID origID;
        if (p.second.first == 0) {
            origID = reg->storeOrdinaryGAtom(atom);
        }
        else {
            origID = reg->storeOrdinaryNAtom(atom);
        }

        // store unfounded atom
        atom.kind = ID::MAINKIND_ATOM;
        if (p.second.first == 0) atom.kind |= ID::SUBKIND_ATOM_ORDINARYG; else atom.kind |= ID::SUBKIND_ATOM_ORDINARYN;
        atom.tuple[0] = unfoundedPredicates[p.first].second;
        ID unfoundedID;
        if (p.second.first == 0) {
            unfoundedID = reg->storeOrdinaryGAtom(atom);
        }
        else {
            unfoundedID = reg->storeOrdinaryNAtom(atom);
        }

        // store shadow atom
        atom.kind = ID::MAINKIND_ATOM;
        if (p.second.first == 0) atom.kind |= ID::SUBKIND_ATOM_ORDINARYG; else atom.kind |= ID::SUBKIND_ATOM_ORDINARYN;
        atom.tuple[0] = p.second.second;
        ID shadowID;
        if (p.second.first == 0) {
            shadowID = reg->storeOrdinaryGAtom(atom);
        }
        else {
            shadowID = reg->storeOrdinaryNAtom(atom);
        }
        DBGLOG(DBG, "Using shadow predicate for " << p.first << " which is " << p.second.second << " and unfounded predicate which is " << unfoundedPredicates[p.first].second);

        // for each shadow atom, either the original atom or the notfounded atom is derived
        {
            Rule r(ID::MAINKIND_RULE | ID::SUBKIND_RULE_REGULAR | ID::PROPERTY_RULE_DISJ);
            r.head.push_back(origID);
            r.head.push_back(unfoundedID);
            r.body.push_back(shadowID);
            idb.push_back(reg->storeRule(r));
        }
    }
}


DLVHEX_NAMESPACE_END

// vim:expandtab:ts=4:sw=4:
// mode: C++
// End:
